% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_and_reporting.R
\name{umxAPA}
\alias{umxAPA}
\alias{summaryAPA}
\title{Creates nicely formatted journal style summaries of lm models, p-values, data-frames etc.}
\usage{
umxAPA(obj = .Last.value, se = NULL, std = FALSE, digits = 2,
  use = "complete", min = 0.001, addComparison = NA,
  report = c("markdown", "html"), lower = TRUE, test = c("Chisq",
  "LRT", "Rao", "F", "Cp"), SEs = TRUE, means = TRUE)
}
\arguments{
\item{obj}{A model (e.g. \link{lm}, lme, glm, t-test), beta-value, or data.frame}

\item{se}{If obj is a beta, se treated as standard-error (returning a CI). 
If obj is a model, used to select effect of interest (blank for all effects). 
Finally, set se to the CI c(lower, upper), to back out the SE.}

\item{std}{Whether to report std betas (re-runs model on standardized data).}

\item{digits}{How many digits to round output.}

\item{use}{If obj is a data.frame, how to handle NAs (default = "complete")}

\item{min}{For a p-value, the smallest value to report numerically (default .001)}

\item{addComparison}{For a p-value, whether to add "</=" default (NA) adds "<" if necessary}

\item{report}{What to return (default = 'markdown'). Use 'html' to open a web table.}

\item{lower}{Whether to not show the lower triangle of correlations for a data.frame (Default TRUE)}

\item{test}{If obj is a glm, which test to use to generate p-values options = "Chisq", "LRT", "Rao", "F", "Cp"}

\item{SEs}{Whether or not to show correlations with their SE (Default TRUE)}

\item{means}{Whether or not to show means in a correlation table (Default TRUE)}
}
\value{
- string
}
\description{
umxAPA creates summaries from a range of inputs. Use it for reporting lm models, effects, and summarizing data.

1. Given an lm, will return a formatted effect, including 95\% CI 
in square brackets, for one of the effects (specified by name in se). e.g.:
\code{\link{umxAPA}}(m1, "wt") yields:

\eqn{\beta} = -5.344 [-6.486, -4.203], p< 0.001

2. Given a dataframe, summaryAPA will return a table of correlations, with
the mean and SD of each variable as the last row. So, 
\code{umxAPA(mtcars[,c("cyl", "wt", "mpg", )])} yields a table of 
correlations, means and SDs thus:

\tabular{lccc}{
        \tab cyl         \tab  wt          \tab mpg          \cr
cyl     \tab 1           \tab  0.78        \tab -0.85        \cr
wt      \tab 0.78        \tab  1           \tab -0.87        \cr
mpg     \tab -0.85       \tab  -0.87       \tab 1            \cr
mean_sd \tab 6.19 (1.79) \tab  3.22 (0.98) \tab 20.09 (6.03)
}

3. Given obj and se, umxAPA returns a CI based on 1.96 times the se.

4. Given only a number as obj will be treated as a p-value as returned in APA format.
}
\examples{

# ========================================
# = Report lm (regression/anova) results =
# ========================================
umxAPA(lm(mpg ~ wt + disp, mtcars)) # All parameters
umxAPA(lm(mpg ~ wt + disp, mtcars), "disp") # Just disp effect
umxAPA(lm(mpg ~ wt + disp, mtcars), std = TRUE) # Standardize effects

# glm example
df = mtcars
df$mpg_thresh = 0
df$mpg_thresh[df$mpg>16] = 1
m1 = glm(mpg_thresh ~ wt + gear,data = df, family = binomial)
umxAPA(m1)

# A t-Test
m1 = t.test(1:10, y = c(7:20))
umxAPA(m1)

# ========================================================
# = Summarize a DATA FRAME: Correlations + Means and SDs =
# ========================================================
umxAPA(mtcars[,1:3])
umxAPA(mtcars[,1:3], digits = 3)
umxAPA(mtcars[,1:3], lower = FALSE)
\dontrun{
umxAPA(mtcars[,1:3], report = "html")
}

# ===============================================
# = CONFIDENCE INTERVAL text from effect and se =
# ===============================================
umxAPA(.4, .3) # parameter 2 interpreted as SE

# Input beta and CI, and back out the SE
umxAPA(-0.030, c(-0.073, 0.013), digits = 3)

# ====================
# = Format a p-value =
# ====================
umxAPA(.0182613)
umxAPA(.000182613)
umxAPA(.000182613,  addComparison=FALSE)

# ========================
# = report a correlation =
# ========================
data(twinData)
selDVs = c("wt1", "wt2")
mzData <- subset(twinData, zygosity \%in\% c("MZFF", "MZMM"))
x = cor.test(~ wt1 + wt2, data = mzData)
umxAPA(x)

}
\references{
- \url{https://github.com/tbates/umx}, \url{https://my.ilstu.edu/~jhkahn/apastats.html}
}
\seealso{
Other Reporting Functions: \code{\link{loadings.MxModel}},
  \code{\link{tmx_is.identified}}, \code{\link{umxEval}},
  \code{\link{umxFactorScores}},
  \code{\link{umxGetParameters}}, \code{\link{umxReduce}},
  \code{\link{umxSummary}}, \code{\link{umxWeightedAIC}},
  \code{\link{umx_APA_model_CI}},
  \code{\link{umx_APA_pval}}, \code{\link{umx_aggregate}},
  \code{\link{umx_names}}, \code{\link{umx_parameters}},
  \code{\link{umx_print}},
  \code{\link{umx_show_fit_or_comparison}},
  \code{\link{umx_show}}, \code{\link{umx_time}},
  \code{\link{umx}}
}
\concept{Reporting Functions}
